<?php

/**
 * Controller for a mock network service.
 *
 * @author kato
 */
class MockNetworkService_Controller {
    private $conversation = array();
    private $conversationLineSeparator = "\n--\n";
    private $process;
    private $pipeToProgramInput;
    private $pipeFromProgramOutput;
    private $pipeFromProgramErrorOutput;
    private $outputString = '';
    private $exitStatus;

    /**
     * Adds a regular expression to match against next set of input from client.
     *
     * Use parentheses to specify subpatterns to use for response placeholders.
     *
     * Use named subpatterns of format (?P<name>...) to save matches
     * for use in later responses.
     *
     * @return MockNetworkService_Controller this instance for call chaining.
     */
    function addExpectation($pattern)
    {
        $this->addEmptyResponseIfMissing();
        $this->conversation[] = $pattern;
        return $this;
    }

    private function addEmptyResponseIfMissing()
    {
        if (0 === (count($this->conversation) % 2))
        {
            $this->conversation[] = '';
        }
    }

    /**
     * Add a response string to send to client.
     *
     * Respose string may contain placeholders to be filled in from earlier
     * matches of client input against expectation patterns.
     *
     * Numeric placeholders of form ${0} will be replaced with matches from
     * latest input only.
     *
     * Named placeholders such as ${name} will be replaced with latest matches
     * with like names.
     *
     * @return MockNetworkService_Controller this instance for call chaining.
     */
    function addResponse($responseString)
    {
        $this->addUnlimitedExpectationIfMissing();
        $this->conversation[] = $responseString;
        return $this;
    }

    private function addUnlimitedExpectationIfMissing()
    {
        if (1 === (count($this->conversation) % 2))
        {
            $this->conversation[] = '/.*/s';
        }
    }

    /**
     * Starts mock network service. It is assumed you have explained it what
     * kind of conversation it should hold with a connecting client.
     *
     * @param $host string address to listen at, default 127.0.0.1
     * @param $port integer number of port to listen at, default 1337
     * @param $verbose boolean whether to report all traffic
     */
    function launch($host = '127.0.0.1', $port = 1337, $verbose = false)
    {
        $this->chooseConversationLineSeparator();
        $command = $this->formCommandString($host, $port, $verbose);
        $this->startProcess($command);
        $this->passConversationToMockNetworkService();
        $this->waitForServerToAnnounceListening();
    }

    private function waitForServerToAnnounceListening()
    {
        fread($this->pipeFromProgramOutput, 8000);
    }

    private function chooseConversationLineSeparator()
    {
        $wholeConversation = implode('', $this->conversation);
        return $this->findGoodConversationLineSeparatorFor($wholeConversation);
    }

    private function findGoodConversationLineSeparatorFor($wholeConversation)
    {
        $i = 0;
        while ($this->conversationLineSeparatorAppearsIn($wholeConversation))
        {
            $this->conversationLineSeparator =
                "\n--NextMockConversationLine--" . ++$i . "--\n";
        }
    }

    private function conversationLineSeparatorAppearsIn($string)
    {
        return false !== strstr($string, $this->conversationLineSeparator);
    }

    private function formCommandString($host, $port, $verbose)
    {
        return
            $this->formStaticCommandString()
            . $this->formCommandOptionString($host, $port, $verbose);
    }

    private function formStaticCommandString()
    {
        return 'php ' . dirname(__FILE__) . '/run.php';
    }

    private function formCommandOptionString($host, $port, $verbose)
    {
        return
            " --host=$host --port=$port"
            . ' --conversationLineSeparator='
            . trim($this->conversationLineSeparator)
            . ($verbose ? ' --verbose' : '');
    }

    private function startProcess($command)
    {
        $descriptors = array(array('pipe', 'r'), array('pipe', 'w'), array('pipe', 'w'));

        $this->process = proc_open('exec ' . $command, $descriptors, $pipes);

        $this->pipeToProgramInput = $pipes[0];
        $this->pipeFromProgramOutput = $pipes[1];
        $this->pipeFromProgramErrorOutput = $pipes[2];
    }

    private function passConversationToMockNetworkService()
    {
        $conversationString =
            implode($this->conversationLineSeparator, $this->conversation);
        fwrite($this->pipeToProgramInput, $conversationString);
        fclose($this->pipeToProgramInput);
    }

    /**
     * See if the mock service has been started and has not finished yet.
     *
     * @return boolean whether the service is currently running.
     */
    function isRunning()
    {
        $status = $this->getStatus();
        return $status['running'];
    }

    /**
     * See if the mock service has finished
     * with conversation completed as planned.
     *
     * @return boolean whether the mock service has exited successfully.
     */
    function hasSucceeded()
    {
        $this->getStatus();
        return $this->exitStatus && ! $this->exitStatus['exitcode'];
    }

    /**
     * See if the mock service has finished
     * because the conversation has not proceeded as planned.
     *
     * @return boolean whether the mock service has exited unsuccessfully.
     */
    function hasFailed()
    {
        $this->getStatus();
        return $this->exitStatus && $this->exitStatus['exitcode'];
    }

    /**
     * Get explanation of what went wrong.
     *
     * @return string report from the mock service.
     */
    function getFailureMessage()
    {
        return $this->outputString ? $this->outputString :
                "No output from MockNetworkService\n";
    }

    /**
     * Get exit code of the mock service process.
     *
     * @return integer exit code of the mock service.
     */
    function getExitCode()
    {
        $this->getStatus();
        if ($this->exitStatus)
        {
            return $this->exitStatus['exitcode'];
        }
    }

    /**
     * Query status of mock network service process.
     *
     * @return array of process status data.
     */
    function getStatus()
    {
        if ($this->exitStatus)
        {
            return $this->exitStatus;
        }
        return $this->getFreshStatusAndFinishIfProcessDone();
    }

    private function getFreshStatusAndFinishIfProcessDone()
    {
        $status = proc_get_status($this->process);
        if ($status['running'])
        {
            $this->collectCurrentOutput();
        }
        else
        {
            $this->finishWithProcess($status);
        }
        return $status;
    }

    private function finishWithProcess($status)
    {
        $this->setExitStatus($status);
        $this->collectAllOutput();
        $this->closeFinishedProcess();
    }

    private function setExitStatus($status)
    {
        $this->exitStatus = $status;
    }

    private function collectCurrentOutput()
    {
        $this->setOutputBlocking(false);
        $this->collectOutput();
    }

    private function collectAllOutput()
    {
        $this->setOutputBlocking(true);
        $this->collectOutput();
    }

    private function setOutputBlocking($onOrOff)
    {
        stream_set_blocking($this->pipeFromProgramOutput, $onOrOff);
        stream_set_blocking($this->pipeFromProgramErrorOutput, $onOrOff);
    }

    private function collectOutput()
    {
        while( strlen($in = fread($this->pipeFromProgramOutput, 500000)) ||
               strlen($in = fread($this->pipeFromProgramErrorOutput, 500000)) )
        {
            $this->outputString .= $in;
        }
    }

    private function closeFinishedProcess()
    {
        proc_close($this->process);
        $this->process = null;
    }

    /**
     * Finish the mock network service process if it is still running.
     */
    function end()
    {
        if ($this->process)
        {
            proc_terminate($this->process);
            do
            {
                $this->getFreshStatusAndFinishIfProcessDone();
            }
            while (!$this->exitStatus && sleep(1) === 0);
        }
    }
}
