<?php
/**
 * MockNetworkService accepts a connection from a client over TCP.
 * Then it holds a conversation with that client following a predefined script.
 *
 * Conversation is specified as an array holding responses and expectations
 * in turn, beginning with a response to send right after connection.
 *
 * Inputs expectations are matched as regular expressions against actual input.
 * Named subpatterns from them will be saved for use in later responses.
 *
 * Responses are strings.
 * Numeric placeholders in them such as ${0} are replaced
 * with matches from the last input.
 * Named placeholders such as ${name} are replaced from saved subpatterns.
 *
 * Service ends succesfully without output at the end of the conversation script.
 *
 * Service exits with an error message and failure status (7) if the client
 * says anything unexpected.
 *
 * @author kato
 */
class MockNetworkService_Server {

    private $listeningSocket;
    private $connectedSocket;
    private $conversation;
    private $matches = array();
    private $conversationStep = 1;
    private $verbose;

    /**
     * Immediately starts a new service with given parameters.
     *
     * @param $host address to serve at
     * @param $port number of port to serve at
     * @param $conversation array of responses and expectations
     * @param $verbose whether to output debugging information
     */
    function __construct($host, $port, $conversation, $verbose)
    {
        $this->verbose = $verbose;
        $this->conversation = $conversation;
        $this->createListeningSocket($host, $port);
    }

    private function createListeningSocket($host, $port)
    {
        $socket =
            socket_create(AF_INET, SOCK_STREAM, getprotobyname('tcp'));
        socket_set_option($socket, SOL_SOCKET, SO_REUSEADDR, 1);
        socket_bind($socket, $host, $port);
        socket_listen($socket);
        $this->listeningSocket = $socket;
    }

    /**
     * Accepts connections and holds the predefined conversation.
     *
     * Exits with a nonzero status upon unexpected input from client.
     */
    function serve()
    {
        echo "Listening\n";
        do
        {
            $this->handleIncomingTraffic();
        }
        while(! empty($this->conversation));
        $this->end();
        $this->debug("Ok.");
    }

    private function handleIncomingTraffic()
    {
        $readableSockets = array($this->getCurrentSocket());
        $null = null;

        $this->reportWaiting();

        socket_select($readableSockets, $null, $null, $null);

        foreach ($readableSockets as $socket)
        {
            $this->handleReadableSocket($socket);
        }
    }

    private function reportWaiting()
    {
        if ($this->verbose)
        {
            if (!$this->connectedSocket)
            {
                $this->debug('Waiting for connection');
            }
            else
            {
                $expected =
                    empty($this->conversation) ? '' : $this->conversation[0];
                $this->debug(
                    "Waiting for #$this->conversationStep: '$expected'\n");
            }
        }
    }

    private function getCurrentSocket()
    {
        return
            $this->connectedSocket ?
                $this->connectedSocket
                : $this->listeningSocket;
    }

    private function handleReadableSocket($socket)
    {
        switch ($socket)
        {
            case $this->listeningSocket:
                $this->connectedSocket = socket_accept($socket);
                $this->debug("Got connection: $this->connectedSocket");
                $this->respond();
            case $this->connectedSocket:
                $this->converse();
        }
    }

    protected function respond($matches = array())
    {
        $response = array_shift($this->conversation);
        foreach (array_merge($matches, $this->matches) as $i => $match)
        {
            $response = preg_replace("/\\\$\{$i}/", $match, $response);
        }
        if (strlen($response))
        {
            $this->debug("Sending response: '$response'");
            socket_write($this->connectedSocket, $response);
        }
    }

    protected function converse()
    {
        $data = $this->readFromOrCloseClientConnection();

        while(strlen($data))
        {
            $data = $this->consumeSomeReceivedDataAndRespondAsAppropriate($data);
        }
    }

    private function readFromOrCloseClientConnection()
    {
        $data = socket_read($this->connectedSocket, 8888);
        $len = strlen($data);
        if (!$len)
        {
            $this->debug("Client closed connection.");
            $this->closeConnectedSocket();
            $expected =
                $this->getCurrentExpectationFor('<CLIENT CLOSED CONNECTION>');
            if ($expected !== '')
            {
                $this->fail("Client disconnected while expected: '$expected'");
            }
            array_shift($this->conversation);
            ++$this->conversationStep;
        }
        else
        {
            $this->debug("Read $len bytes");
        }
        return $data;
    }

    private function consumeSomeReceivedDataAndRespondAsAppropriate($data)
    {
        $expected = $this->getCurrentExpectationFor($data);

        if (preg_match($expected, $data, $matches))
        {
            return $this->handleMatchedConversationStep($data, $matches);
        }
        $safeData = preg_replace(
                '/[^\x20-\x7E]/e',
                'sprintf("<%02X>", ord("$0"))',
                $data);
        $this->fail("Expected '$expected' but got '$safeData'");
    }

    private function getCurrentExpectationFor($data)
    {
        if (empty($this->conversation))
        {
            $this->fail("Received input past end of plan: '$data'");
        }
        return $this->conversation[0];
    }

    private function handleMatchedConversationStep($data, $matches)
    {
        array_shift($this->conversation);
        $data = $this->consumeMatchedData($data, $matches);
        $this->saveNamedMatches($matches);
        $this->respond($matches);
        ++$this->conversationStep;
        return $data;
    }

    private function consumeMatchedData($data, $matches)
    {
        $this->debug("Matched #$this->conversationStep: '$matches[0]'");
        $oldLen = strlen($data);
        $len = strlen($matches[0]);
        return $len < $oldLen ? substr($data, $len) : '';
    }

    protected function saveNamedMatches($matches)
    {
        foreach($matches as $key => $value)
        {
            if(!is_int($key))
            {
                $this->matches[$key] = $value;
            }
        }
    }

    protected function fail($message)
    {
        echo $message . "\n";
        $this->end();
        exit(7);
    }

    private function end()
    {
        $this->closeConnectedSocket();
        socket_close($this->listeningSocket);
    }

    private function closeConnectedSocket()
    {
        if ($this->connectedSocket)
        {
            socket_close($this->connectedSocket);
            $this->connectedSocket = null;
        }
    }

    private function debug($message)
    {
        if ($this->verbose)
        {
            echo "$message\n";
        }
    }
}
